"""Based on https://github.com/cheminfo/chemcalc"""

import re
import warnings
from typing import Any
import pandas as pd
import polars as pl
import numpy as np
from IsoSpecPy import ParseFormula
from mascope_tools.composition.heuristic_filter import (
    apply_heuristic_rules,
    match_isotopic_pattern,
)
from mascope_tools.composition import utils
from mascope_tools.composition.models import (
    SearchContext,
    Atom,
    Result,
    CompositionFinderWarning,
)
from mascope_tools.composition.constants import (
    UNSATURATION_COEFFICIENTS,
)


def _is_notebook():
    try:
        from IPython import get_ipython

        shell = get_ipython().__class__.__name__
        return shell == "ZMQInteractiveShell"
    except Exception:
        return False


if _is_notebook():
    from tqdm.notebook import tqdm
else:
    from tqdm import tqdm


def assign_compositions(
    peaks: pd.DataFrame, params: dict, heuristic_params: dict | None = None
) -> pd.DataFrame:
    """Assign molecular compositions to a set of peaks.

    :param peaks: DataFrame with 'mz' and 'intensity' columns.
    :type peaks: pd.DataFrame
    :param params: Parameters for composition finding.
    :type params: dict
    :param heuristic_params: Parameters for heuristic filtering.
    :type heuristic_params: dict, optional
    :return: A DataFrame with assigned compositions and related information.
    :rtype: pd.DataFrame
    """
    # Convert peaks to Polars DataFrame
    peaks = pl.from_pandas(peaks).sort("mz")
    peak_height_threshold = params.get("peak_height_threshold", 0.0)
    peaks_to_match = peaks.filter(pl.col("intensity") >= peak_height_threshold)
    peaks_to_match = peaks_to_match.sort("mz")

    mzs = peaks_to_match["mz"].to_numpy()
    results_per_peak, assigned_mzs, mass_log_messages = [], set(), {}

    for mz in tqdm(mzs, desc="Assigning compositions..."):
        if mz in assigned_mzs:
            continue
        params["monoisotopic_mass"] = mz
        params["target_monoisotopic_mass"] = mz

        compositions = find_compositions(params)
        comp_results = compositions.get("results", [])
        all_candidates = (
            ", ".join([r["formula"] for r in comp_results[1:]])
            if len(comp_results) > 1
            else ""
        )

        if comp_results:
            candidates, log_messages = apply_heuristic_rules(
                comp_results, params=heuristic_params
            )
            mass_log_messages[mz] = log_messages
            if candidates:
                candidates, all_matched_isotopes = match_isotopic_pattern(
                    candidates, peaks
                )
            else:
                all_matched_isotopes = []
            if not candidates:
                results_per_peak.append(
                    {
                        "formula": "---",
                        "ion": "---",
                        "mz": mz,
                        "other_candidates": all_candidates,
                        "isotope_label": "---",
                    }
                )
                continue
            main_candidate = candidates[0].copy()
            main_candidate["mz"] = mz
            main_candidate["formula"] = main_candidate.get("formula", "---")
            main_candidate["other_candidates"] = all_candidates

            if all_matched_isotopes:
                isotopic_results, assigned_mzs = process_isotopes(
                    main_candidate, all_matched_isotopes, assigned_mzs
                )
                results_per_peak.extend(isotopic_results)
            else:
                # No isotopic pattern matched, just add the main result
                main_candidate["isotope_label"] = "M0"
                results_per_peak.append(main_candidate)
                assigned_mzs.add(main_candidate["mz"])

    unmatched_peaks = set(mzs) - assigned_mzs
    for mz in unmatched_peaks:
        results_per_peak.append(
            {
                "mz": mz,
                "formula": "---",
                "ion": "---",
                "isotope_label": "---",
                "other_candidates": "",
            }
        )

    matches = pd.DataFrame(results_per_peak)
    # --- Format results --- #
    matches = matches.sort_values(by=["mz", "mz_error_ppm"])
    # Drop duplicate m/z entries, keeping the one with the lowest mz_error_ppm
    matches = matches.drop_duplicates(subset=["mz"], keep="first")
    matches = sort_matches_by_formula(matches)
    # Add isotope label to ion string
    matches = update_ion_with_isotope_label(matches)
    # Show mz, formula, ion, isotope_label and then all other columns
    first_columns = ["mz", "formula", "ion", "isotope_label", "ionization_mechanism"]
    cols = first_columns + [col for col in matches.columns if col not in first_columns]
    matches = matches[cols].reset_index(drop=True)

    return matches, mass_log_messages


def process_isotopes(
    main_candidate: dict, all_matched_isotopes: list, assigned_mzs: set
) -> tuple:
    """Process and add isotopic pattern results

    :param main_candidate: Most likely composition for the monoisotopic m/z and related data.
    :type main_candidate: dict
    :param all_matched_isotopes: List of matched isotopic patterns.
    :type all_matched_isotopes: list
    :param assigned_mzs: The m/z values that have already been assigned to a composition.
    :type assigned_mzs: set
    :return: A tuple containing:
        - List of results per peak including isotopic variants.
        - Updated set of assigned m/z values.
    :rtype: tuple
    """
    results_per_peak = []
    # Take the first matched isotopic pattern (best scoring)
    matched_isotopes = all_matched_isotopes[0]
    isotope_mzs = matched_isotopes["masses"]
    isotope_labels = matched_isotopes["labels"]
    isotope_pred_mzs = matched_isotopes["predicted_masses"]
    isotope_pred_ints = matched_isotopes["predicted_intensities"]
    isotope_mz_errors = matched_isotopes["mass_errors_ppm"]
    isotope_intensity_errors = matched_isotopes["intensity_errors"]
    if isotope_mzs[0] != 0:
        # Extract and process base peak (M0)
        m0_mass = isotope_mzs[0]
        main_candidate["mz"] = m0_mass
        main_candidate["observed_mass"] = m0_mass
        main_candidate["predicted_mz"] = isotope_pred_mzs[0]
        main_candidate["predicted_intensity"] = isotope_pred_ints[0]
        main_candidate["isotope_label"] = "M0"
        main_candidate["mz_error_ppm"] = isotope_mz_errors[0]
        main_candidate["intensity_error"] = isotope_intensity_errors[0]
        results_per_peak.append(main_candidate)
        assigned_mzs.add(m0_mass)

        # Extract and process higher isotopes
        for idx in range(1, len(isotope_mzs)):
            iso_mz = isotope_mzs[idx]
            if iso_mz == 0:
                continue
            if iso_mz in assigned_mzs:
                continue
            iso_result = main_candidate.copy()
            iso_result["mz"] = iso_mz
            iso_result["observed_mass"] = iso_mz
            iso_result["isotope_label"] = isotope_labels[idx]
            iso_result["predicted_mz"] = isotope_pred_mzs[idx]
            iso_result["predicted_intensity"] = isotope_pred_ints[idx]
            iso_result["mz_error_ppm"] = isotope_mz_errors[idx]
            iso_result["intensity_error"] = isotope_intensity_errors[idx]
            iso_result["neutral_mass"] = iso_result["neutral_mass"] + (iso_mz - m0_mass)
            results_per_peak.append(iso_result)
            assigned_mzs.add(iso_mz)

    return results_per_peak, assigned_mzs


def find_compositions(params: dict[str, Any]) -> dict:
    """Find molecular compositions based on the provided parameters.

    :param params: Parameters for the composition search.
    :type params: dict[str, Any]
    :return: A dictionary containing the search results and related information.
    :rtype: dict
    """
    ctx = SearchContext()
    ctx.build(params)
    ctx.atoms = utils.parse_atom_count_ranges(ctx.element_count_ranges)
    ctx.atoms.sort(key=lambda a: a.mass)

    ctx.min_inner_mass, ctx.max_inner_mass = calc_min_max_inner_mass(ctx.atoms)

    ionization_mech_string_list = get_ionization_mech_string_list(params)

    target_mz = ctx.target_monoisotopic_mass
    ctx.abs_tolerance = target_mz * ctx.mass_range * 1e-6  # ppm -> Da around target m/z

    all_results: list[Result] = []

    for ionization_mech_string in ionization_mech_string_list:
        ionization_mech = utils.parse_ionization(ionization_mech_string)

        # Ion shift: ion m/z = neutral_mass + ion_shift
        ion_shift = (
            ionization_mech.mass if ionization_mech.addition else -ionization_mech.mass
        )
        required_neutral_mass = (
            target_mz - ion_shift
        )  # what neutral mass would give the target m/z

        # Ionization peak case: no analyte mass (neutral mass ~ 0)
        if abs(required_neutral_mass) <= ctx.abs_tolerance:
            ion_charge = "+" if ionization_mech.charge > 0 else "-"
            ion_formula = ionization_mech.formula + ion_charge
            all_results.append(
                Result(
                    formula="Ionization peak",
                    neutral_mass=0.0,
                    composition_error_ppm=0.0,
                    unsaturation=None,
                    ion=ion_formula,
                    ionization_mechanism=ionization_mech.mascope_notation,
                    observed_mass=target_mz,
                )
            )
            continue

        # Skip mechanisms that would imply negative / zero neutral masses
        if required_neutral_mass <= 0:
            continue

        # Store mechanism info in context for recursion
        ctx.ionization_mechanism = ionization_mech
        ctx.ion_shift = ion_shift
        ctx.results_found = 0  # reset per ionization mechanism

        for res in recursive_search(0, [], 0.0, ctx):
            all_results.append(res)

    all_results.sort(key=lambda r: r.composition_error_ppm)

    return_result_count_only = utils.parse_bool(
        params.get("return_result_count_only", False)
    )
    return_typed_format = utils.parse_bool(params.get("return_typed_format", False))
    if return_result_count_only:
        return {"count": len(all_results)}

    return {
        "results": [format_result(r, return_typed_format) for r in all_results],
        "count": len(all_results),
        "options": params,
    }


def recursive_search(idx: int, counts: list, current_mass: float, ctx: SearchContext):
    """A recursive function to explore all possible combinations of atom counts.

    :param idx: Current index in the list of atoms.
    :type idx: int
    :param counts: Current counts of each atom type.
    :type counts: list
    :param current_mass: Current total mass of the composition.
    :type current_mass: float
    :param ctx: Search context containing parameters and state.
    :type ctx: SearchContext
    :yield: Result objects for valid compositions.
    :rtype: Iterator[Result]
    """
    if ctx.results_found >= ctx.max_result_rows:
        return

    # Evaluate full composition
    if idx == len(ctx.atoms):
        ion_mz = current_mass + ctx.ion_shift
        if abs(ion_mz - ctx.target_monoisotopic_mass) <= ctx.abs_tolerance:
            if ctx.use_unsaturation:
                unsat = get_unsaturation(ctx.atoms, counts)
                if not (ctx.min_unsaturation <= unsat <= ctx.max_unsaturation):
                    return
                if ctx.only_integer_unsaturation and not unsat.is_integer():
                    return
            else:
                unsat = None

            atomic_counts = {
                ctx.atoms[i].symbol: counts[i] for i in range(len(ctx.atoms))
            }
            formula = utils.to_hill_order(atomic_counts)
            ctx.results_found += 1
            ion_formula = utils.combine_formula_and_ionization(
                formula, ctx.ionization_mechanism
            )
            error_ppm = (
                abs(ion_mz - ctx.target_monoisotopic_mass)
                / ctx.target_monoisotopic_mass
                * 1e6
            )
            yield Result(
                formula=formula,
                neutral_mass=current_mass,
                composition_error_ppm=error_ppm,
                unsaturation=unsat,
                ion=ion_formula,
                ionization_mechanism=ctx.ionization_mechanism.mascope_notation,
                observed_mass=ctx.target_monoisotopic_mass,
            )
        return

    atom = ctx.atoms[idx]
    min_inner = ctx.min_inner_mass
    max_inner = ctx.max_inner_mass
    tol = ctx.abs_tolerance
    shift = ctx.ion_shift
    target_mz = ctx.target_monoisotopic_mass

    # Feasible count bounds for this atom (neutral mass domain)
    feasible_min = max(
        atom.min_count,
        int(
            np.ceil(
                ((target_mz - shift) - tol - current_mass - max_inner[idx]) / atom.mass
            )
        )
        - 1,
    )
    feasible_max = min(
        atom.max_count,
        int(
            np.floor(
                ((target_mz - shift) + tol - current_mass - min_inner[idx]) / atom.mass
            )
        )
        + 1,
    )
    if feasible_min > feasible_max:
        return

    for atom_count in range(feasible_min, feasible_max + 1):
        if ctx.results_found >= ctx.max_result_rows:
            return
        new_mass = current_mass + atom_count * atom.mass

        if idx < len(ctx.atoms) - 1:
            min_mass = new_mass + min_inner[idx]
            max_mass = new_mass + max_inner[idx]
            min_ion = min_mass + shift
            max_ion = max_mass + shift

            # Too heavy already (even minimal remaining mass overshoots)
            if (min_ion - target_mz) > tol:
                break
            # Still too light (even maximal remaining mass below window)
            if (target_mz - max_ion) > tol:
                continue

        yield from recursive_search(idx + 1, counts + [atom_count], new_mass, ctx)


def format_result(r, return_typed_format):
    """Format a Result object into a dictionary."""
    base = r.to_dict()
    base["formula"] = (
        {"type": "formula", "value": r.formula} if return_typed_format else r.formula
    )
    return base


def get_ionization_mech_string_list(params: dict) -> list:
    """Get a list of ionizations from the params dictionary."""
    ionizations = params.get("ionizations", None)
    if ionizations:
        return [ionization for ionization in ionizations.split(",")]
    else:
        raise ValueError("No ionization mechanisms provided.")


def get_neutral_mass_and_ionization_mech(
    target_mass: float, ion: str
) -> tuple[float, str]:
    if ion:
        ionization_mech = utils.parse_ionization(ion)
        if ionization_mech.addition:
            # If it's an addition, we subtract mass
            neutral_mass = target_mass - ionization_mech.mass
        else:
            # If it's a subtraction, we add mass
            neutral_mass = target_mass + ionization_mech.mass
        return neutral_mass, ionization_mech
    return target_mass, None


def calc_min_max_inner_mass(atoms):
    """Prepare suffix arrays of minimal and maximal remaining masses AFTER each index.

    Returns:
        min_suffix[i]: minimal mass contribution of atoms with index > i
        max_suffix[i]: maximal mass contribution of atoms with index > i
        For convenience lengths match len(atoms); min_suffix[-1] == max_suffix[-1] == 0.
    """
    n = len(atoms)
    min_suffix = [0.0] * n
    max_suffix = [0.0] * n
    running_min = 0.0
    running_max = 0.0
    # Build from the end toward the front; suffix after i
    for i in range(n - 1, -1, -1):
        min_suffix[i] = running_min
        max_suffix[i] = running_max
        running_min += atoms[i].min_count * atoms[i].mass
        running_max += atoms[i].max_count * atoms[i].mass
    return min_suffix, max_suffix


def get_unsaturation(atoms: list[Atom], counts: list[int]) -> float:
    """Calculate the unsaturation (double bond equivalents) of a molecular formula.

    Warns if an atom's unsaturation coefficient is not supported.

    :param atoms: Iterable of Atom objects representing the elements in the formula.
    :type atoms: list[Atom]
    :param counts: List of counts for each atom in the formula.
    :type counts: list[int]
    :return: Unsaturation value (double bond equivalents).
    :rtype: float
    """
    unsaturation_value = 0
    for i, atom in enumerate(atoms):
        coefficient = UNSATURATION_COEFFICIENTS.get(atom.symbol, 0)
        if atom.symbol not in UNSATURATION_COEFFICIENTS:
            warnings.warn(
                f"Unsaturation coefficient for '{atom.symbol}' not supported, using {coefficient}.",
                CompositionFinderWarning,
            )
        unsaturation_value += coefficient * counts[i]
    return (unsaturation_value + 2) / 2.0


def _formula_sort_key(formula: str) -> tuple[int, int, str]:
    """
    Generate a sorting key for a chemical formula based on atomic composition.
    Priority:
        0: Only C and H
        1: Only C, H, and O
        2: Only C, H, O, and N
        3: All other C-containing
        4: Non-carbon containing
    """
    if formula == "---":
        return (5, 0, formula)
    atoms = set(ParseFormula(formula).keys())
    if "C" not in atoms:
        return (4, len(atoms), formula)
    if atoms <= {"C", "H"}:
        return (0, len(atoms), formula)
    if atoms <= {"C", "H", "O"}:
        return (1, len(atoms), formula)
    if atoms <= {"C", "H", "O", "N"}:
        return (2, len(atoms), formula)
    return (3, len(atoms), formula)


def sort_matches_by_formula(matches: pd.DataFrame) -> pd.DataFrame:
    """Sort a DataFrame of chemical formulae by atomic composition:
    1. C,H only
    2. C,H,O only
    3. C,H,O,N only
    4. Other C-containing
    5. Non-carbon containing
    Within each group, sort by number of atoms, then lexicographically.

    :param matches: Dataframe with matched peaks
    :type matches: pd.DataFrame
    :return: Sorted matches
    :rtype: pd.DataFrame
    """
    sort_keys = matches["formula"].apply(_formula_sort_key)
    return (
        matches.assign(_sort_key=sort_keys)
        .sort_values("_sort_key")
        .drop("_sort_key", axis=1)
        .reset_index(drop=True)
    )


def replace_atom_with_isotope(ion: str, isotope_label: str) -> str:
    """
    Replaces one atom of an element in a chemical formula with a specified isotope.
    """
    if not isinstance(isotope_label, str) or isotope_label in {"M0", "---", ""}:
        return ion
    isotope_label = f"[{isotope_label}]"

    element_match = re.search(r"\[\d*([A-Z][a-z]*)\]", isotope_label)
    if not element_match:
        raise ValueError("Invalid isotope format. Expected format like '[13C]'.")
    isotope_element = element_match.group(1)

    # Separate the charge at the end of the formula, if any
    charge = ion[-1] if ion[-1] in "+-" else ""
    ion = ion[:-1] if charge else ion

    element_counts = ParseFormula(ion)

    # 4. Check if the isotope's element exists in the formula
    if isotope_element not in element_counts:
        raise ValueError(
            f"Element '{isotope_element}' not found in the formula '{ion}'."
        )

    # Decrement the count of the target element
    element_counts[isotope_element] -= 1

    # Rebuild the formula string
    new_formula_parts = [isotope_label]
    for element in element_counts.keys():
        count = element_counts[element]
        if count == 0:
            continue  # Skip elements with a count of zero
        elif count == 1:
            new_formula_parts.append(element)
        else:
            new_formula_parts.append(f"{element}{count}")

    # Append the charge and join everything into the final string
    return "".join(new_formula_parts) + charge


def update_ion_with_isotope_label(matches: pd.DataFrame) -> pd.DataFrame:
    """
    Update the 'ion' column in the matches DataFrame to prepend the isotope label
    (if present and not 'M0' or '---') before the first element symbol.
    """
    matches = matches.copy()
    matches["ion"] = [
        replace_atom_with_isotope(ion, isotope_label)
        for ion, isotope_label in zip(matches["ion"], matches["isotope_label"])
    ]
    return matches



from typing import Any
import os
import pickle
import hashlib
import threading


class CacheManager:
    """
    Manages disk caching of Python objects using pickle serialization.
    Each object is associated with a unique key, which is hashed to create
    a filename in the specified cache directory.
    """

    def __init__(self, cache_dir: str) -> None:
        self.cache_dir = os.path.abspath(cache_dir)
        os.makedirs(self.cache_dir, exist_ok=True)
        self._lock = threading.Lock()

    def _key_to_path(self, key: Any) -> str:
        key_bytes = pickle.dumps(key)
        digest = hashlib.sha256(key_bytes).hexdigest()
        return os.path.join(self.cache_dir, digest)

    def save(self, key: Any, obj: Any) -> None:
        path = self._key_to_path(key)
        tmp_path = path + ".__tmp__"
        with self._lock, open(tmp_path, "wb") as f:
            pickle.dump(obj, f, protocol=pickle.HIGHEST_PROTOCOL)
        os.replace(tmp_path, path)

    def load(self, key: Any) -> Any | None:
        path = self._key_to_path(key)
        try:
            with self._lock, open(path, "rb") as f:
                return pickle.load(f)
        except Exception:
            return None

    def exists(self, key: Any) -> bool:
        path = self._key_to_path(key)
        return os.path.exists(path)

    def invalidate(self, key: Any) -> None:
        path = self._key_to_path(key)
        try:
            os.remove(path)
        except FileNotFoundError:
            pass



ELECTRON_MASS = 5.48579909065e-4

DEFAULT_MAXIMUM_ROWS = 100
DEFAULT_SEARCH_ELEMENT_COUNT_RANGES = "C0-100 H0-200 N0-20 O0-20"
# Default mass range threshold to look for formulas in ppm
DEFAULT_MASS_RANGE_THRESHOLD_PPM = 1

DEFAULT_MAXIMUM_UNSATURATION = 50.0
ISOTOPE_ABUNDANCE_THRESHOLD = 0.01
ISOTOPE_MATCHING_MZ_TOLERANCE_PPM = 5
ISOTOPE_MATCHING_INTENSITY_TOLERANCE = 0.4


# Wiley spectral database:
# H/C ratio 0.1...6.0 in 99.99% of all formulas.
DEFAULT_ELEMENTAL_RATIO_RANGE = {
    "H/C": (0.1, 6.0),
    "N/C": (0.0, 2.0),
    "O/C": (0.0, 2.0),
    "S/C": (0.0, 0.1),
    "P/C": (0.0, 0.05),
    "Cl/C": (0.0, 0.05),
    "Br/C": (0.0, 0.05),
    "F/C": (0.0, 0.05),
    "I/C": (0.0, 0.05),
}

UNSATURATION_COEFFICIENTS = {
    "C": 2,
    "H": -1,
    "N": 1,
    "O": 0,
    "F": -1,
    "Cl": -1,
    "Br": -1,
    "I": -1,
    "C[13]": 2,
    "O[18]": 0,
}



from dataclasses import dataclass, asdict
from mascope_tools.composition.constants import (
    DEFAULT_MAXIMUM_UNSATURATION,
    DEFAULT_MAXIMUM_ROWS,
    DEFAULT_SEARCH_ELEMENT_COUNT_RANGES,
    DEFAULT_MASS_RANGE_THRESHOLD_PPM,
)
from mascope_tools.composition import utils


class CompositionFinderException(Exception):
    pass


class CompositionFinderWarning(UserWarning):
    pass


class HeuristicRuleWarning(UserWarning):
    pass


@dataclass
class Atom:
    symbol: str
    min_count: int
    max_count: int
    mass: float


@dataclass
class IonizationMechanism:
    mascope_notation: str
    addition: bool
    formula: str
    charge: int
    mass: float


@dataclass
class SearchContext:
    atoms: list[Atom] = None
    min_inner_mass: float = None
    max_inner_mass: float = None
    neutral_mass: float = None
    mass_range: float = None
    use_unsaturation: bool = False
    min_unsaturation: float = 0.0
    max_unsaturation: float = DEFAULT_MAXIMUM_UNSATURATION
    only_integer_unsaturation: bool = False
    max_result_rows: int = DEFAULT_MAXIMUM_ROWS
    ionization_mechanism: IonizationMechanism | None = None
    ion_shift: float = 0.0
    target_monoisotopic_mass: float = None
    results_found: int = 0

    def build(self, params: dict):
        self.mass_range = float(
            params.get("mass_range_ppm", DEFAULT_MASS_RANGE_THRESHOLD_PPM)
        )
        self.max_result_rows = int(params.get("max_result_rows", DEFAULT_MAXIMUM_ROWS))
        self.element_count_ranges = params.get(
            "element_count_ranges", DEFAULT_SEARCH_ELEMENT_COUNT_RANGES
        )
        self.min_unsaturation = float(params.get("min_unsaturation", 0))
        self.max_unsaturation = float(params.get("max_unsaturation", 50))
        self.only_integer_unsaturation = utils.parse_bool(
            params.get("only_integer_unsaturation", False)
        )
        self.use_unsaturation = utils.parse_bool(params.get("use_unsaturation", False))
        self.target_monoisotopic_mass = float(
            params.get(
                "monoisotopic_mass", params.get("target_monoisotopic_mass", "-1")
            )
        )


@dataclass
class Result:
    formula: str
    neutral_mass: float
    composition_error_ppm: float
    ion: str | None
    ionization_mechanism: str | None
    observed_mass: float
    unsaturation: float | None = None
    other_candidates: list[str] | None = None

    def to_dict(self):
        d = asdict(self)
        # Remove None values for optional fields
        return {k: v for k, v in d.items() if v is not None}



"""
Based on 7 Golden Rules by https://bmcbioinformatics.biomedcentral.com/articles/10.1186/1471-2105-8-105
"""

from typing import Any
import numpy as np
from scipy.spatial.distance import cosine
import polars as pl
from IsoSpecPy import IsoThreshold, ParseFormula, PeriodicTbl
from mascope_tools.composition.constants import (
    DEFAULT_ELEMENTAL_RATIO_RANGE,
    ISOTOPE_ABUNDANCE_THRESHOLD,
    ELECTRON_MASS,
    ISOTOPE_MATCHING_MZ_TOLERANCE_PPM,
    ISOTOPE_MATCHING_INTENSITY_TOLERANCE,
)

# Limit isotopic matching to the most plausible candidates
ISOTOPE_CANDIDATE_LIMIT = 64


def rule_element_ratio(
    candidates: pl.DataFrame, **kwargs
) -> tuple[pl.Series, list[str]]:
    log_messages = []
    params = kwargs.get("params", {})
    carbon_element_ratio_range = params.get(
        "carbon_element_ratio_range", DEFAULT_ELEMENTAL_RATIO_RANGE
    )
    non_carbon_element_ratio_range = params.get("non_carbon_element_ratio_range", {})

    # Early return if no candidates
    if candidates.is_empty():
        return pl.Series([], dtype=pl.Boolean), log_messages

    formulas = candidates.get_column("formula").to_list()

    # Parse all formulas once and convert to a structured format
    counts_list = [ParseFormula(f) for f in formulas]

    # Get all unique elements across all formulas
    all_elements = set()
    for counts in counts_list:
        all_elements.update(counts.keys())
    all_elements = sorted(all_elements)

    # Create a matrix where rows are formulas and columns are elements
    n_formulas = len(counts_list)
    n_elements = len(all_elements)
    element_matrix = np.zeros((n_formulas, n_elements), dtype=np.int32)

    # Fill the matrix
    element_to_idx = {elem: idx for idx, elem in enumerate(all_elements)}
    for i, counts in enumerate(counts_list):
        for elem, count in counts.items():
            element_matrix[i, element_to_idx[elem]] = count

    # Determine which formulas have carbon
    carbon_idx = element_to_idx.get("C")
    has_carbon = carbon_idx is not None and element_matrix[:, carbon_idx] > 0

    # Initialize mask - all True initially
    final_mask = np.ones(n_formulas, dtype=bool)

    def apply_ratio_rules_vectorized(ratio_range, apply_to_mask):
        """Apply ratio rules using vectorized operations"""
        if not ratio_range or not np.any(apply_to_mask):
            return np.ones(n_formulas, dtype=bool)

        rule_mask = np.ones(n_formulas, dtype=bool)

        for ratio, (min_val, max_val) in ratio_range.items():
            num, denom = ratio.split("/")

            # Get indices for numerator and denominator elements
            num_idx = element_to_idx.get(num)
            denom_idx = element_to_idx.get(denom)

            if num_idx is None or denom_idx is None:
                continue

            # Get counts for numerator and denominator
            num_counts = element_matrix[:, num_idx]
            denom_counts = element_matrix[:, denom_idx]

            # Only apply rule where both elements exist and denominator > 0
            has_both_elements = (num_counts > 0) & (denom_counts > 0)
            applicable_mask = apply_to_mask & has_both_elements

            if not np.any(applicable_mask):
                continue

            # Calculate ratios only where applicable (avoid division by zero)
            ratios = np.full(n_formulas, np.inf)
            ratios[applicable_mask] = (
                num_counts[applicable_mask] / denom_counts[applicable_mask]
            )

            # Check if ratios are within bounds
            ratio_valid = (ratios >= min_val) & (ratios <= max_val)

            # Update rule mask: pass if not applicable OR ratio is valid
            rule_mask &= np.logical_not(applicable_mask) | ratio_valid

        return rule_mask

    # Apply carbon-specific ratios to formulas with carbon
    if np.any(has_carbon) and carbon_element_ratio_range:
        carbon_mask = apply_ratio_rules_vectorized(
            carbon_element_ratio_range, has_carbon
        )
        final_mask &= carbon_mask

    # Apply non-carbon ratios to formulas without carbon
    no_carbon = np.logical_not(has_carbon)
    if np.any(no_carbon) and non_carbon_element_ratio_range:
        non_carbon_mask = apply_ratio_rules_vectorized(
            non_carbon_element_ratio_range, no_carbon
        )
        final_mask &= non_carbon_mask

    return pl.Series(final_mask), log_messages


def rule_valence(candidates: pl.DataFrame, **kwargs) -> tuple[pl.Series, list[str]]:
    """Valence rules (even/odd electron)."""
    # TODO: requires charge and electron count info
    mask = pl.Series([True] * candidates.height)
    log_messages = []
    return mask, log_messages  # Placeholder, always returns True


def rule_senior(candidates: pl.DataFrame, **kwargs) -> tuple[pl.Series, list[str]]:
    """Senior's rules (structural feasibility)."""
    # TODO: requires graph theory/structure generation
    mask = pl.Series([True] * candidates.height)
    log_messages = []
    return mask, log_messages  # Placeholder, always returns True


def rule_known_chemical_space(
    candidates: pl.DataFrame, **kwargs
) -> tuple[pl.Series, list[str]]:
    """Known chemical space (database matching)."""
    # TODO: requires access to some chemical database
    mask = pl.Series([True] * candidates.height)
    log_messages = []
    return mask, log_messages  # Placeholder, always returns True


# From lightweight to heavyweight, these rules are applied in order.
HEURISTIC_RULES = [
    rule_element_ratio,
    rule_valence,
    rule_senior,
    rule_known_chemical_space,
]


def apply_heuristic_rules(
    candidates: list[dict[str, Any]],
    params: dict[str, Any] = None,
) -> list[dict[str, Any]]:
    """
    Filter candidate formulas using the heuristic rules.
    Returns only those that pass all rules.

    :param candidates: List of candidate formula dicts (or Result objects).
    :return: Filtered list of candidates.
    """
    if params is None:
        params = {}
    log_messages = []
    candidates = pl.DataFrame(candidates)
    if candidates.is_empty():
        log_messages.append("No candidates provided for heuristic filtering.")
        return [], log_messages

    if "Ionization peak" in candidates.get_column("formula").to_list():
        # Skip all rules for ionization peaks
        return (
            candidates.filter(pl.col("formula") == "Ionization peak").to_dicts(),
            log_messages,
        )

    for i, rule in enumerate(HEURISTIC_RULES):
        if candidates.is_empty():
            log_messages.append(
                f"No candidates from passed the rule: {HEURISTIC_RULES[i-1].__name__}"
            )
            break
        rule_mask, rule_log_messages = rule(candidates, params=params)
        log_messages.extend(rule_log_messages)
        candidates = candidates.filter(rule_mask)

    return candidates.to_dicts(), log_messages


def match_isotopic_pattern(
    candidates: list[dict[str, Any]], peaks: pl.DataFrame
) -> tuple[list[dict[str, Any]], list[dict[str, np.ndarray | list[str]]]]:
    """Matches isotopic patterns against candidates.

    :param candidates: List of candidate formula dicts.
    :type candidates: list[dict[str, Any]]
    :param peaks: Sorted dataframe of peaks with 'mz' and 'intensity' columns.
    :type peaks: pl.DataFrame
    :return: Tuple of filtered candidates, and a list of isotope data dicts (per candidate).
    :rtype: tuple[list[dict[str, Any]], list[dict[str, np.ndarray | list[str]]]]
    """
    mzs = peaks["mz"].to_numpy()
    intensities = peaks["intensity"].to_numpy()

    candidates_df = pl.DataFrame(candidates)
    if candidates_df.is_empty():
        candidates_df = candidates_df.with_columns(
            pl.lit(0.0, dtype=pl.Float64).alias("isotopic_pattern_score")
        )
        return candidates_df.to_dicts(), []

    # Keep only the most promising candidates for heavy work
    candidates_df = candidates_df.sort("composition_error_ppm").head(
        ISOTOPE_CANDIDATE_LIMIT
    )

    # If ionization peak: skip isotopic matching and return score 1.0
    if "Ionization peak" in candidates_df.get_column("formula").to_list():
        candidates_df = candidates_df.with_columns(
            pl.lit(1.0, dtype=pl.Float64).alias("isotopic_pattern_score")
        )
        return candidates_df.to_dicts(), []

    ion_formulas, ion_charges = _extract_formulae_and_charges(
        candidates_df.get_column("ion")
    )

    scores = np.zeros(candidates_df.height, dtype=float)
    all_isotope_data = []

    for ind, (ion_formula, ion_charge) in enumerate(zip(ion_formulas, ion_charges)):
        predicted_mzs, predicted_intensities, isotope_labels = predict_isotopes(
            ion_formula, ion_charge
        )
        is_isotope_predicted = len(predicted_mzs) > 0
        if not is_isotope_predicted:
            all_isotope_data.append(
                {
                    "masses": [],
                    "mass_errors_ppm": [],
                    "intensity_errors": [],
                    "labels": [],
                    "predicted_masses": [],
                    "predicted_intensities": [],
                }
            )
            continue

        observed_masses = np.zeros_like(predicted_mzs)
        observed_intensities = observed_masses.copy()
        observed_mass_errors_ppm = observed_masses.copy()
        observed_intensity_error = observed_masses.copy()

        # Normalize predicted intensities relative to monoisotopic (base) peak
        predicted_rel = predicted_intensities / predicted_intensities[0]

        base_peak_intensity = None
        for i, p_mz in enumerate(predicted_mzs):
            mz_delta = p_mz * ISOTOPE_MATCHING_MZ_TOLERANCE_PPM * 1e-6
            mz_min, mz_max = p_mz - mz_delta, p_mz + mz_delta

            start_idx = np.searchsorted(mzs, mz_min, side="left")
            end_idx = np.searchsorted(mzs, mz_max, side="right")
            no_peaks_in_window = start_idx >= end_idx

            if no_peaks_in_window:
                continue

            window_mzs = mzs[start_idx:end_idx]
            window_intensities = intensities[start_idx:end_idx]
            if not window_mzs.size:
                continue

            matched_index = np.argmin(np.abs(window_mzs - p_mz))
            matched_mz = window_mzs[matched_index]
            matched_intensity = window_intensities[matched_index]
            is_base_peak = i == 0

            if is_base_peak:
                base_peak_intensity = matched_intensity
                observed_intensities[0] = matched_intensity
                observed_masses[0] = matched_mz
                observed_mass_errors_ppm[0] = abs(matched_mz - p_mz) / p_mz * 1e6
                observed_intensity_error[0] = 0.0
                continue  # move to next isotope

            # Require monoisotopic established before evaluating higher isotopes
            if base_peak_intensity is None or base_peak_intensity == 0:
                continue

            predicted_rel_intensity = predicted_rel[i]
            observed_rel_intensity = matched_intensity / base_peak_intensity
            intensity_error = (
                abs(predicted_rel_intensity - observed_rel_intensity)
                / predicted_rel_intensity
            )

            if intensity_error <= ISOTOPE_MATCHING_INTENSITY_TOLERANCE:
                observed_intensities[i] = matched_intensity
                observed_masses[i] = matched_mz
                observed_mass_errors_ppm[i] = abs(matched_mz - p_mz) / p_mz * 1e6
                observed_intensity_error[i] = intensity_error

        scores[ind] = score_pattern(
            observed_masses,
            observed_mass_errors_ppm,
            observed_intensities,
            observed_intensity_error,
            predicted_rel,
        )

        matched_isotopes = {
            "masses": observed_masses,
            "mass_errors_ppm": observed_mass_errors_ppm,
            "intensity_errors": observed_intensity_error,
            "labels": isotope_labels,
            "predicted_masses": predicted_mzs,
            "predicted_intensities": predicted_rel,
        }

        all_isotope_data.append(matched_isotopes)

    candidates_df = candidates_df.with_columns(
        pl.Series(values=scores, name="isotopic_pattern_score")
    ).sort("isotopic_pattern_score", descending=True)

    score_sorted_indices = np.argsort(scores)[::-1]
    all_isotope_data = [all_isotope_data[i] for i in score_sorted_indices]

    return candidates_df.to_dicts(), all_isotope_data


def predict_isotopes(
    ion_formula: str, ion_charge: int
) -> tuple[np.ndarray, np.ndarray, list[str]]:
    """Predict isotopic pattern for a given ion formula and charge.

    :param ion_formula: Ion formula string.
    :type ion_formula: str
    :param ion_charge: Ion charge (e.g., +1, -1).
    :type ion_charge: int
    :return: Tuple of predicted m/z values, relative intensities, and isotope labels.
    :rtype: tuple[np.ndarray, np.ndarray, list[str]]
    """
    try:
        predicted_peaks = IsoThreshold(
            formula=ion_formula,
            threshold=ISOTOPE_ABUNDANCE_THRESHOLD,
            get_confs=True,
        )
        predicted_masses_neutral = np.fromiter(predicted_peaks.masses, dtype=float)
        predicted_intensities = np.fromiter(predicted_peaks.probs, dtype=float)

        composition = ParseFormula(ion_formula)
        elements = list(composition.keys())
        elemental_masses = [PeriodicTbl.symbol_to_masses[el] for el in elements]
        isotope_labels = [
            conf_to_label(conf, elements, elemental_masses)
            for conf in predicted_peaks.confs
        ]
        # Convert neutral masses to m/z
        predicted_mzs = (predicted_masses_neutral - ELECTRON_MASS * ion_charge) / abs(
            ion_charge
        )
    except Exception:
        predicted_mzs, predicted_intensities, isotope_labels = [], [], []

    return predicted_mzs, predicted_intensities, isotope_labels


def score_pattern(
    observed_masses: np.ndarray,
    observed_mass_errors_ppm: np.ndarray,
    observed_intensities: np.ndarray,
    observed_intensity_error: np.ndarray,
    predicted_rel: np.ndarray,
) -> float:
    """
    Scores the match between observed and predicted isotopic patterns.
    Returns a score between 0 and 1, where 1 is a perfect match.
    """
    # Require monoisotopic detection
    if observed_intensities[0] > 0:
        observed_rel_intensities = observed_intensities / observed_intensities[0]
        matched_peaks_count = np.sum(observed_masses > 0)

        # 1. Pattern scoring
        cosine_dist = cosine(predicted_rel, observed_rel_intensities)
        pattern_score = 1 - cosine_dist if not np.isnan(cosine_dist) else 0.0

        # 2. Intensity scoring
        total_intensity_error = np.sum(observed_intensity_error)
        avg_intensity_error = (
            total_intensity_error / matched_peaks_count
            if matched_peaks_count > 0
            else ISOTOPE_MATCHING_INTENSITY_TOLERANCE
        )
        intensity_score = max(
            0, 1 - (avg_intensity_error / ISOTOPE_MATCHING_INTENSITY_TOLERANCE)
        )

        # 3. Mass Accuracy Score
        total_mass_error_ppm = np.sum(observed_mass_errors_ppm)
        avg_mass_error = (
            total_mass_error_ppm / matched_peaks_count
            if matched_peaks_count > 0
            else ISOTOPE_MATCHING_MZ_TOLERANCE_PPM
        )
        mass_score = max(0, 1 - (avg_mass_error / ISOTOPE_MATCHING_MZ_TOLERANCE_PPM))

        # 4. Combined score.
        # pattern_score and intensity_score get lower weights because they are less reliable,
        # we may have only base peak detected.
        score = 0.2 * pattern_score + 0.2 * intensity_score + 0.6 * mass_score
    else:
        score = 0.0

    return score


def conf_to_label(conf, elements, isotope_masses):
    """Return isotope label string.

    :param conf: isotope counts for each element in the formula.
    :type conf: list[list[int]]
    :param elements: list of elements in the formula.
    :type elements: list[str]
    :param isotope_masses: list of isotope masses for each element.
    :type isotope_masses: list[list[float]]
    """
    label_parts = []
    for el, iso_counts, iso_masses in zip(elements, conf, isotope_masses):
        for idx, count in enumerate(iso_counts):
            if count == 0:
                continue

            # For the most abundant isotope (usually index 0), skip label unless it's the only one (M0)
            if idx == 0:
                continue

            mass_number = int(round(iso_masses[idx]))

            label_parts.append(f"{mass_number}{el}{count if count > 1 else ''}")

    if not label_parts:
        return "M0"
    return "+".join(label_parts)


def _extract_formulae_and_charges(ions: pl.Series) -> tuple[list[str], list[int]]:
    """Extracts formulae and charges from ion strings

    :param ions: Array of ion strings.
    :type ions: pl.Series
    :return: Tuple of lists containing ion formulas and their charges.
    :rtype: tuple[list[str], list[int]]
    """
    ions_arr = ions.to_numpy().astype(str)
    # Get last character for each ion string
    last_chars = np.array([s[-1] if len(s) >= 1 else "" for s in ions_arr])
    # Check if last char is + or -
    is_charged = np.isin(last_chars, ["+", "-"])
    # Remove last char if charged, else keep as is
    ion_formulas = [
        s[:-1] if charged else s for s, charged in zip(ions_arr, is_charged)
    ]
    # Assign charge: +1 for '+', -1 for '-', else 1
    ion_charges = [1 if c == "+" else -1 if c == "-" else 1 for c in last_chars]
    return ion_formulas, ion_charges



"""
Implementation of the virtual lock mass (VLM) algorithm based on https://doi.org/10.1038/s41598-019-44923-8
"""

from abc import ABC, abstractmethod
import numpy as np
import pandas as pd
import heapq
from collections import deque
import warnings


class Spectrum(ABC):
    """Abstract base class for a mass spectrum"""

    @property
    @abstractmethod
    def mz(self) -> np.ndarray:
        pass

    @property
    @abstractmethod
    def intensity(self) -> np.ndarray:
        pass

    @property
    @abstractmethod
    def resolution(self) -> np.ndarray:
        pass

    @property
    @abstractmethod
    def metadata(self) -> dict:
        pass


class CentroidedSpectrum(Spectrum):
    def __init__(self, mz, intensity, resolution, signal_to_noise, metadata=None):
        self._mz = np.array(mz)
        self._intensity = np.array(intensity)
        self._resolution = np.array(resolution)
        self._signal_to_noise = np.array(signal_to_noise)
        self._metadata = metadata or {}

    @property
    def mz(self):
        return self._mz

    @property
    def intensity(self):
        return self._intensity

    @property
    def resolution(self):
        return self._resolution

    @property
    def signal_to_noise(self):
        return self._signal_to_noise

    @property
    def metadata(self):
        return self._metadata


class Spectra:
    def __init__(self, spectra: list[Spectrum], timestamps: np.ndarray):
        self.spectra = spectra
        self.timestamps = pd.to_datetime(timestamps, unit="s")

        non_empty_spectra = [s for s in spectra if s.mz.size > 0]
        if not non_empty_spectra:
            self.mz_min, self.mz_max = 0, 0
        else:
            self.mz_min = min(s.mz.min() for s in non_empty_spectra)
            self.mz_max = max(s.mz.max() for s in non_empty_spectra)

    def __getitem__(self, index: int) -> Spectrum:
        return self.spectra[index]

    def __len__(self) -> int:
        return len(self.spectra)

    def __iter__(self):
        return iter(self.spectra)

    def _cluster_and_map_peaks(self, window_factor: float) -> pd.DataFrame:
        """
        A private helper method that contains the core logic for flattening,
        sorting, and clustering peaks from all spectra.

        Returns a pandas DataFrame mapping every single peak to a final cluster.
        """
        if not self.spectra:
            return pd.DataFrame()

        # 1. Concatenate data from all spectra, keeping track of the original scan index
        all_data = []
        for i, s in enumerate(self.spectra):
            for j in range(len(s.mz)):
                all_data.append(
                    (i, s.mz[j], s.intensity[j], s.resolution[j], s.signal_to_noise[j])
                )

        if not all_data:
            return pd.DataFrame()

        # Create a DataFrame and sort by m/z for efficient processing
        df = pd.DataFrame(
            all_data,
            columns=["scan_idx", "mz", "intensity", "resolution", "signal_to_noise"],
        )
        df.sort_values("mz", inplace=True)
        df.reset_index(drop=True, inplace=True)

        # 2. Iteratively find clusters and assign a unique cluster_id to each peak
        cluster_id = 0
        current_index = 0
        cluster_assignments = np.full(len(df), -1, dtype=int)

        while current_index < len(df):
            # Define cluster window based on the current peak's measured resolution
            start_mz = df.at[current_index, "mz"]
            start_resolution = df.at[current_index, "resolution"]

            # Peak width (FWHM) = m/z / R
            peak_width = start_mz / start_resolution if start_resolution > 0 else 0
            delta = peak_width * window_factor

            # Find all peaks within the window
            end_index = df["mz"].searchsorted(start_mz + delta, side="right")

            # Assign the current cluster_id to all peaks in this window
            cluster_assignments[current_index:end_index] = cluster_id

            cluster_id += 1
            current_index = end_index

        df["cluster_id"] = cluster_assignments
        return df

    def compute_sum_spectrum(self, window_factor=1, average=False) -> "Spectrum":
        """
        Compute sum spectrum, optionally averaged.
        """
        clustered_df = self._cluster_and_map_peaks(window_factor)
        if clustered_df.empty:
            return CentroidedSpectrum(
                mz=[], intensity=[], resolution=[], signal_to_noise=[], metadata={}
            )

        # Group by the cluster_id to aggregate peaks
        grouped = clustered_df.groupby("cluster_id")

        # Calculate total intensity, SNR, and weighted averages for m/z and resolution
        total_intensity = grouped["intensity"].sum().values

        weighted_avg_mz = grouped.apply(
            lambda g: np.average(g["mz"], weights=g["intensity"])
        )
        weighted_avg_res = grouped.apply(
            lambda g: np.average(g["resolution"], weights=g["intensity"])
        )

        def compute_total_snr(g):
            S = g["intensity"].values
            R = g["signal_to_noise"].values
            numerator = np.sum(S)
            denominator = np.sqrt(np.sum((S / R) ** 2))
            return numerator / denominator if denominator > 0 else 0.0

        total_snr = grouped.apply(compute_total_snr).values

        if average and len(self.spectra) > 0:
            total_intensity /= len(self.spectra)
            total_snr /= len(self.spectra)

        return CentroidedSpectrum(
            mz=weighted_avg_mz.values,
            intensity=total_intensity,
            resolution=weighted_avg_res.values,
            signal_to_noise=total_snr,
            metadata={"average": average, "window_factor": window_factor},
        )

    def get_timeseries(self, window_factor=1) -> pd.DataFrame:
        """
        Returns a DataFrame with timestamps as columns and aggregated m/z as indices.
        """
        clustered_df = self._cluster_and_map_peaks(window_factor)
        if clustered_df.empty:
            return pd.DataFrame()

        # Use pandas pivot_table for an efficient transformation
        timeseries_df = clustered_df.pivot_table(
            index="cluster_id",
            columns="scan_idx",
            values="intensity",
            aggfunc="sum",
            fill_value=0.0,
        )

        # Calculate the representative m/z for each cluster_id to use as the final index
        weighted_avg_mz = clustered_df.groupby("cluster_id").apply(
            lambda g: np.average(g["mz"], weights=g["intensity"])
        )

        # Set the DataFrame index to the m/z values and columns to timestamps
        timeseries_df.index = weighted_avg_mz
        timeseries_df.columns = self.timestamps[timeseries_df.columns]
        timeseries_df.index.name = "m/z"

        return timeseries_df


class MassAligner:
    """
    Identifies and applies mass correction based on the algorithms described in
    "Mass spectra alignment using virtual lock-masses" by Brochu et al.

    This class implements the heap-based detection (Algorithm 4), overlap
    deletion (Algorithm 5), and single-pass correction (Algorithm 6).
    """

    def __init__(
        self,
        window_factor: float,
        min_peak_intensity: float = 0.0,
        min_fraction: float = 1.0,
    ):
        """
        :param window_factor: Factor to determine the window size for alignment points.
                              This is multiplied by the FWHM of the peaks.
        :param min_peak_intensity: Minimum intensity for a peak to be considered.
        :param min_fraction: Minimum fraction of spectra that must contain a peak for it
                             to be considered a point. Set to 1.0 for strict VLMs.
        """
        self.window_factor = window_factor
        self.min_peak_intensity = min_peak_intensity
        self.min_fraction = min_fraction
        self.points_mz: np.ndarray | None = None
        self.points_resolution: np.ndarray | None = None

    def fit(self, spectra: Spectra) -> np.ndarray:
        """
        Finds VLM or Alignment points in the spectra.
        This method implements Algorithm 4 from the paper.

        :param spectra: Spectra object to analyze.
        :return: Array of m/z values for the found points.
        """
        num_spectra = len(spectra)
        min_matches = int(np.ceil(num_spectra * self.min_fraction))

        # 1. Filter spectra by intensity and prepare for processing
        filtered_spectra_list = []
        for s in spectra:
            mask = s.intensity >= self.min_peak_intensity
            filtered_spectra_list.append(
                (s.mz[mask], s.intensity[mask], s.resolution[mask])
            )

        # 2. Initialize data structures from the paper
        heap = _Heap(filtered_spectra_list)
        active_seq = _ActiveSequence(num_spectra, self.window_factor, min_matches)

        found_point_mzs = []
        found_point_resolutions = []
        found_flag = False

        # 3. Main loop of Algorithm 4
        while not heap.empty():
            if active_seq.is_valid(heap):
                found_flag = True

            if not active_seq.insert(heap):
                if found_flag:
                    found_point_mzs.append(active_seq.calculate_average_mz())
                    found_point_resolutions.append(
                        active_seq.calculate_average_resolution()
                    )
                    found_flag = False
                active_seq.advance_lower_bound()

        # Final check after heap is exhausted
        while active_seq.peak_deque:
            if active_seq.is_valid(heap):
                found_point_mzs.append(active_seq.calculate_average_mz())
                found_point_resolutions.append(
                    active_seq.calculate_average_resolution()
                )
                break
            active_seq.advance_lower_bound()

        # 4. Remove overlaps as per Algorithm 5
        unique_mzs, unique_indices = np.unique(found_point_mzs, return_index=True)
        unique_resolutions = np.array(found_point_resolutions)[unique_indices]
        self.points_mz, self.points_resolution = self._delete_overlaps(
            unique_mzs, unique_resolutions, self.window_factor
        )
        return self.points_mz

    def _delete_overlaps(
        self, point_mzs: np.ndarray, point_resolutions: np.ndarray, window_factor: float
    ) -> np.ndarray:
        """Implements Algorithm 5: deleteOverlaps."""
        if point_mzs.size < 2:
            return point_mzs, point_resolutions

        sort_indices = np.argsort(point_mzs)
        sorted_mzs = point_mzs[sort_indices]
        sorted_res = point_resolutions[sort_indices]

        point_fwhms = sorted_mzs / sorted_res
        point_intervals = point_fwhms * self.window_factor

        is_overlapping = np.zeros(sorted_mzs.shape, dtype=bool)

        # Vectorized overlap check
        end_i = sorted_mzs[:-1] + point_intervals[:-1]
        start_next = sorted_mzs[1:] - point_intervals[1:]
        overlaps = end_i >= start_next

        is_overlapping[:-1][overlaps] = True
        is_overlapping[1:][overlaps] = True

        non_overlapping_original_indices = sort_indices[~is_overlapping]
        return (
            point_mzs[non_overlapping_original_indices],
            point_resolutions[non_overlapping_original_indices],
        )

    def transform(self, spectra: Spectra) -> Spectra:
        """
        Applies correction to spectra based on fitted points.
        This method implements Algorithm 6 from the paper.

        :param spectra: Spectra object to correct.
        :return: A new Spectra object with corrected m/z values.
        """
        if self.points_mz is None or len(self.points_mz) < 2:
            warnings.warn(
                "Correction requires at least 2 alignment points. Returning original spectra."
            )
            return spectra

        corrected_spectra_list = [self._apply_correction(s) for s in spectra]
        return Spectra(corrected_spectra_list, spectra.timestamps)

    def _apply_correction(self, spectrum: Spectrum) -> Spectrum:
        """Applies correction to a single spectrum."""
        mz = spectrum.mz
        if mz.size == 0 or self.points_mz is None or self.points_mz.size < 2:
            return spectrum

        # 1. Find insertion points for alignment points in the spectrum's m/z array
        insertion_indices = np.searchsorted(mz, self.points_mz)

        # Handle cases where insertion index is out of bounds
        insertion_indices = np.clip(insertion_indices, 1, len(mz) - 1)

        # 2. For each alignment point, the closest peak is either at index i or i-1
        dist1 = np.abs(mz[insertion_indices] - self.points_mz)
        dist2 = np.abs(mz[insertion_indices - 1] - self.points_mz)

        # 3. Choose the index corresponding to the smaller distance
        is_prev_closer = dist2 < dist1
        observed_indices = insertion_indices - is_prev_closer.astype(int)
        min_dists = np.where(is_prev_closer, dist2, dist1)

        # 4. Filter out bad matches where the closest peak is still too far away
        point_fwhms = self.points_mz / self.points_resolution
        max_allowed_dist = point_fwhms * self.window_factor

        valid_match_mask = min_dists <= max_allowed_dist
        observed_indices[~valid_match_mask] = -1

        # --- Perform linear interpolation with filters ---
        corrected_mz = np.copy(mz)
        for i in range(len(self.points_mz) - 1):
            idx_start = observed_indices[i]
            idx_end = observed_indices[i + 1]

            if idx_start == -1 or idx_end == -1:
                continue

            mz_obs_start = mz[idx_start]
            mz_obs_end = mz[idx_end]

            if mz_obs_end <= mz_obs_start:
                continue

            vlm_true_start = self.points_mz[i]
            vlm_true_end = self.points_mz[i + 1]

            slope = (vlm_true_end - vlm_true_start) / (mz_obs_end - mz_obs_start)

            intercept = vlm_true_start - slope * mz_obs_start

            mask = (np.arange(len(mz)) >= idx_start) & (np.arange(len(mz)) <= idx_end)
            corrected_mz[mask] = slope * mz[mask] + intercept

        return CentroidedSpectrum(
            mz=corrected_mz,
            intensity=spectrum.intensity,
            resolution=spectrum.resolution,
            signal_to_noise=spectrum.signal_to_noise,
            metadata=spectrum.metadata,
        )


def calibrate_aligned_spectra(
    spectra: Spectra,
    target_mz: float,
    tol_ppm: float = 10.0,
) -> tuple[Spectra, list[float]]:
    """
    Calibrates spectra using a proportional (multiplicative) one-point correction.

    :param spectra: Collection of spectra to calibrate.
    :type spectra: Spectra
    :param target_mz: The m/z value to align peaks to.
    :type target_mz: float
    :param tol_ppm: Tolerance in ppm for matching m/z values, defaults to 10.0.
    :type tol_ppm: float, optional
    :return: A tuple containing:
        - A new Spectra object with calibrated m/z values.
        - A list of correction factors applied to each spectrum.
    :rtype: tuple[Spectra, list[float]]
    :raises ValueError: If no peaks are found within the specified tolerance.
    """
    delta = target_mz * tol_ppm * 1e-6
    calibrated_spectra = []
    correction_factors = []
    for spectrum in spectra:
        mask = np.abs(spectrum.mz - target_mz) <= delta
        if not np.any(mask):
            raise ValueError(
                f"No peaks found within {tol_ppm} ppm of target m/z {target_mz}."
            )

        observed_mz = spectrum.mz[mask]
        observed_intensity = spectrum.intensity[mask]

        max_intensity_index = np.argmax(observed_intensity)
        observed_lock_mass = observed_mz[max_intensity_index]

        correction_factors.append(target_mz / observed_lock_mass)

        calibrated_spectrum = CentroidedSpectrum(
            mz=spectrum.mz * correction_factors[-1],
            intensity=spectrum.intensity,
            resolution=spectrum.resolution,
            signal_to_noise=spectrum.signal_to_noise,
            metadata=spectrum.metadata,
        )
        calibrated_spectra.append(calibrated_spectrum)

    return Spectra(calibrated_spectra, spectra.timestamps), correction_factors


class _Heap:

    def __init__(self, spectra_list: list[tuple[np.ndarray, np.ndarray, np.ndarray]]):
        self._heap = []
        self._spectra_list = spectra_list
        # For each spectrum, track the index of the next peak to be added
        self._next_peak_indices = [0] * len(self._spectra_list)
        self._init_heap()

    def _init_heap(self):
        """Initializes the heap with the first peak of each spectrum."""
        for spec_idx, spectrum in enumerate(self._spectra_list):
            if spectrum[0].size > 0:
                # Push (m/z, spectrum_index, resolution) to the heap
                mz = spectrum[0][0]
                resolution = spectrum[0][2]
                heapq.heappush(self._heap, (mz, spec_idx, resolution))
                self._next_peak_indices[spec_idx] = 1

    def top(self) -> tuple[float, int, float] | None:
        """Returns the (m/z, spec_idx, resolution) of the peak at the top of the heap."""
        return self._heap[0] if not self.empty() else None

    def pop_and_feed(self) -> tuple[float, int, float] | None:
        """Removes the top peak and adds the next peak from the same spectrum. Returns (m/z, spec_idx, resolution)."""
        if self.empty():
            return None

        # Pop the lowest m/z peak from the heap
        mz, spec_idx, resolution = heapq.heappop(self._heap)

        # Feed the next peak from the same spectrum, if available
        next_idx = self._next_peak_indices[spec_idx]
        spectrum_mzs = self._spectra_list[spec_idx][0]
        spectrum_resolutions = self._spectra_list[spec_idx][2]
        if next_idx < len(spectrum_mzs):
            heapq.heappush(
                self._heap,
                (spectrum_mzs[next_idx], spec_idx, spectrum_resolutions[next_idx]),
            )
            self._next_peak_indices[spec_idx] += 1

        return (mz, spec_idx, resolution)

    def empty(self) -> bool:
        return not self._heap


class _ActiveSequence:
    """Implements the 'active sequence' data structure A from the paper."""

    def __init__(self, num_spectra: int, window_factor: float, min_peaks: int):
        self.peak_deque = (
            deque()
        )  # Doubly linked list for peaks (mz, spec_idx, resolution)
        self.spectrum_flag = [
            False
        ] * num_spectra  # Boolean vector to track spectra in peak_deque
        self.last_removed_mz = -1.0  # m/z of the last removed peak
        self.average_mz = 0.0  # Average m/z of peaks in peak_deque
        self.average_resolution = 0.0  # Average resolution of peaks in peak_deque
        self.window_factor = window_factor  # Times of FWHM for the window
        self.m = num_spectra  # Total number of spectra
        self.min_peaks = min_peaks  # Minimum peaks for a valid sequence

    def calculate_average_mz(self) -> float:
        return self.average_mz

    def calculate_average_resolution(self) -> float:
        return self.average_resolution

    def is_valid(self, heap: "_Heap") -> bool:
        """Checks if peak_deque is a valid VLM/Alignment sequence (Algorithm 1 modified)."""
        if len(self.peak_deque) < self.min_peaks:
            return False

        # This is the key modification for Alignment Points vs. VLMs.
        # For strict VLMs, this would be `len(self.peak_deque) != self.m`.
        # By using min_peaks, we allow for fractional presence.
        if self.min_peaks == self.m and len(self.peak_deque) != self.m:
            return False

        average_FWHM = (
            self.average_mz / self.average_resolution
            if self.average_resolution > 0
            else 0.0
        )

        # Check if the next peak from the heap is isolated
        if (
            not heap.empty()
            and heap.top()[0] <= self.average_mz + average_FWHM * self.window_factor
        ):
            return False

        # Check if all peaks in peak_deque are within the window of the average
        if self.peak_deque[-1][0] > self.average_mz + average_FWHM * self.window_factor:
            return False
        if self.peak_deque[0][0] < self.average_mz - average_FWHM * self.window_factor:
            return False

        # Check if the last removed peak is outside the window
        if self.last_removed_mz >= self.average_mz - average_FWHM * self.window_factor:
            return False

        return True

    def advance_lower_bound(self):
        """Removes the first peak from peak_deque (Algorithm 2)."""
        if not self.peak_deque:
            return
        mz, spec_idx, resolution = self.peak_deque.popleft()
        self.spectrum_flag[spec_idx] = False
        self.last_removed_mz = mz

        # Recalculate the average
        if self.peak_deque:
            old_mz_sum = self.average_mz * (len(self.peak_deque) + 1)
            self.average_mz = (old_mz_sum - mz) / len(self.peak_deque)

            old_resolution_sum = self.average_resolution * (len(self.peak_deque) + 1)
            self.average_resolution = (old_resolution_sum - resolution) / len(
                self.peak_deque
            )
        else:
            self.average_mz = 0.0
            self.average_resolution = 0.0

    def insert(self, heap: "_Heap") -> bool:
        """Tries to insert the top peak from the heap into peak_deque (Algorithm 3)."""
        if heap.empty():
            return False
        top_peak_mz, top_peak_spec_idx, top_peak_resolution = heap.top()

        # If a peak from this spectrum is already in the sequence, cannot add.
        if self.spectrum_flag[top_peak_spec_idx]:
            return False

        # Calculate the new average
        new_average_mz = (self.average_mz * len(self.peak_deque) + top_peak_mz) / (
            len(self.peak_deque) + 1
        )
        new_average_resolution = (
            self.average_resolution * len(self.peak_deque) + top_peak_resolution
        ) / (len(self.peak_deque) + 1)
        new_average_FWHM = (
            new_average_mz / new_average_resolution
            if new_average_resolution > 0
            else 0.0
        )

        # Check if the window condition holds with the new peak
        front_mz = self.peak_deque[0][0] if self.peak_deque else top_peak_mz
        if (top_peak_mz <= new_average_mz + new_average_FWHM * self.window_factor) and (
            front_mz >= new_average_mz - new_average_FWHM * self.window_factor
        ):
            mz, spec_idx, resolution = heap.pop_and_feed()
            self.spectrum_flag[spec_idx] = True
            self.peak_deque.append((mz, spec_idx, resolution))
            self.average_mz = new_average_mz
            self.average_resolution = new_average_resolution
            return True

        return False
